module Functions.Image
  ( Image(..)
  , ImageConversionSettings(..)
  , openImage
  , saveImage
  , convertImage
  ) where

import Types (IsFunction(..), IsFunctionIO(..), Token)
import DependencyGenerator (DepGenM, TokenableTo(..), toTupleToken,
                            runFunction, runFunctionIO, runFunctionIO_)

import qualified Codec.Picture as CP
import qualified Codec.Picture.STBIR as CPS


newtype Image = ImageWrapper (CP.Image CP.PixelRGB8)
  deriving (Eq)

instance Show Image where
  show = const "<image>"

data ImageConversionSettings = ResizeToWidth Int
  deriving (Eq, Show)

data OpenImage = OpenImage deriving Show
instance IsFunctionIO OpenImage FilePath Image where
  evalFunctionIO OpenImage s = do
    imageOrig <- CP.readImage s
    case imageOrig of
      Left e -> error ("unexpected error: " ++ e)
      Right image -> pure $ ImageWrapper $ CP.convertRGB8 image
  functionIOTouchesFilesystem OpenImage = False

openImage :: TokenableTo FilePath a => a -> DepGenM (Token Image)
openImage a = runFunctionIO OpenImage =<< toToken a


data SaveImage = SaveImage deriving Show
instance IsFunctionIO SaveImage (Image, FilePath) () where
  evalFunctionIO SaveImage (ImageWrapper image, s) =
    CP.saveJpgImage 90 s $ CP.ImageRGB8 image
  functionIOTouchesFilesystem SaveImage = True

saveImage :: (TokenableTo Image a, TokenableTo FilePath b) => a -> b -> DepGenM ()
saveImage a b = runFunctionIO_ SaveImage =<< toTupleToken a b


data ConvertImage = ConvertImage deriving Show
instance IsFunction ConvertImage (Image, ImageConversionSettings) Image where
  evalFunction ConvertImage (ImageWrapper image, ResizeToWidth widthResized) =
    let sizeFactor :: Double
        sizeFactor = fromIntegral (CP.imageWidth image) / fromIntegral widthResized
        heightResized = round (fromIntegral (CP.imageHeight image) / sizeFactor)
    in ImageWrapper $ CPS.resize CPS.defaultOptions widthResized heightResized image

convertImage :: (TokenableTo Image a, TokenableTo ImageConversionSettings b) => a -> b -> DepGenM (Token Image)
convertImage a b = runFunction ConvertImage =<< toTupleToken a b
