{-# LANGUAGE GADTs #-}
module Types.Dependency
  ( Action(..)
  , F(..)
  , Dependency(..)
  , makeDependency
  , actionSourceType
  , actionTargetType
  , actionTouchesFilesystem
  , formatDependencyTrees
  ) where

import Types.Token (Token(..))
import Types.Functions (IsFunction(), IsFunctionIO(..))

import Type.Reflection (Typeable, TypeRep, typeRep)
import Text.Printf (printf)
import Data.Text (Text)
import qualified Data.Text as T

data Action a b where
  Function :: IsFunction f a b => f -> Action a b
  InlineFunction :: (Typeable a, Typeable b, Show b) => F a b -> Action a b
  FunctionIO :: IsFunctionIO f a b => f -> Action a b
  Inject :: (Typeable a, Show a) => a -> Action () a
  FilterComp :: (Typeable a, Show a) => Action ([a], [Bool]) [a]
  UntupleFst :: (Typeable a, Show a, Typeable b, Show b) => Action (a, b) a
  UntupleSnd :: (Typeable a, Show a, Typeable b, Show b) => Action (a, b) b
  UnzipFst :: (Typeable a, Show a, Typeable b, Show b) => Action [(a, b)] [a]
  UnzipSnd :: (Typeable a, Show a, Typeable b, Show b) => Action [(a, b)] [b]
  MapComp :: (Typeable a, Show a, Typeable b, Show b) => [Dependency] -> Token a -> Token b -> Action [a] [b]

deriving instance Show (Action a b)

newtype F a b = F (a -> b)

instance Show (F a b) where
  show = const "<function>"

data Dependency where
  Dependency :: TypeRep a -> Token a -> Action a b -> TypeRep b -> Token b -> Dependency
deriving instance Show Dependency

makeDependency :: (Typeable a, Show a, Typeable b, Show b) => Token a -> Action a b -> Token b -> Dependency
makeDependency a action b = Dependency typeRep a action typeRep b

actionSourceType :: Typeable a => Action a b -> TypeRep a
actionSourceType _ = typeRep

actionTargetType :: Typeable b => Action a b -> TypeRep b
actionTargetType _ = typeRep

actionTouchesFilesystem :: Action a b -> Bool
actionTouchesFilesystem = \case
  Function _ -> False
  InlineFunction _ -> False
  FunctionIO f -> functionIOTouchesFilesystem f
  Inject _ -> False
  FilterComp -> False
  UntupleFst -> False
  UntupleSnd -> False
  UnzipFst -> False
  UnzipSnd -> False
  MapComp subDeps _ _ -> any dependencyTouchesFilesystem subDeps
  where dependencyTouchesFilesystem (Dependency _ _ action _ _) =
          actionTouchesFilesystem action

formatDependencyTrees :: [Dependency] -> Text
formatDependencyTrees = T.concat . (formatDependencyTrees' "")
  where formatDependencyTrees' indentation = concatMap (formatDependencyTree indentation)

        formatDependencyTree indentation (Dependency _ a action _ b) =
          concat [ [ indentation ]
                 , formatToken a
                 , [ " -> " ]
                 , formatToken b
                 , [ ": " ]
                 , formatAction indentation action
                 ]

        formatToken :: Token a -> [Text]
        formatToken = \case
          Token i ->
            [ T.pack (printf "%03d" i) ]
          TupleToken a b ->
            concat [ [ "tup(" ]
                   , formatToken a
                   , [ ", " ]
                   , formatToken b
                   , [ ")" ]
                   ]
          ZipToken a b ->
            concat [ [ "zip(" ]
                   , formatToken a
                   , [ ", " ]
                   , formatToken b
                   , [ ")" ]
                   ]
          ListToken ts ->
            [ "["
            , T.intercalate ", " (map (T.concat . formatToken) ts)
            , "]"
            ]
          NoToken ->
            [ "--" ]

        formatAction :: Text -> Action a b -> [Text]
        formatAction indentation = \case
          MapComp subDeps innerInput innerOutput ->
            concat [ [ "MapComp(" ]
                   , formatToken innerInput
                   , [ " -> " ]
                   , formatToken innerOutput
                   , [ "):\n" ]
                   , formatDependencyTrees' (T.append indentation "|  ") subDeps
                   ]
          action ->
            [ T.pack (show action)
            , "\n"
            ]
